-------------localisation-------------
local L = AceLibrary("AceLocale-2.2"):new("HealGraph")

L:RegisterTranslations("enUS", function()
	return {
		["Update interval"] = true,
        ["Set the interval in which the HealGraph updates (in seconds)."] = true,
        ["Create new graph"] = true,
        ["Creates a new graph with the passed id."] = true,
        ["<unit name>"] = true,
        ["<unique graph id>"] = true,
        ["<seconds>"] = true,
        ["Time span"] = true,
        ["Sets the time span the graph should display."] = true,
        ["Remove"] = true,
        ["Removes the graph."] = true,  
        ["Create mark"] = true,
        ["Creates a marking line at the specified time."] = true,
        ["Change to unit"] = true,
        ["Changes the unit of the graph to the passed one. Can be any valid unitid or raidmember name. As well as \"healtarget\" for the unit currently healed."] = true,
        ["Show incoming damage (Experimental)"] = true,
        ["Enables or disables the tracking and display of incoming damage. This is still very experimental and will probably produce some error. You can test it, but don't use it in any serious situation."] = true,
        ["Time span for past"] = true,
        ["Sets the time span lines which already happened should be displayed (the area left from the orange zero-line)."] = true,
        
	}
end)


----------------Libraries----------------
local HealComm = LibStub:GetLibrary("LibHealComm-3.0", true)
local IncEvents = LibStub:GetLibrary("LibIncomingEvents-1.0", true)
local IncEvents2 = LibStub:GetLibrary("MobStats", true)
--local IHL = AceLibrary("IncomingHealsLib-1.0")



-------------create Ace stuff-------------

HealGraph = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")

---------------saved variables------------

HealGraph:RegisterDB("HealGraphDB", "HealGraphDBPC")
HealGraph:RegisterDefaults("profile", {
    UpdateInterval = 0.02,
    FrameTime = 4.0,
    GraphDB = {},
    Graphs = nil,
    EnableIncDmg = false,
    eventColors = nil,
    --hideOutsideRaid = false,
} )

-------------------Options---------------
local options = { 
    type='group',
    args = {        
        updateInterval = {
            type = 'range',
            name = L["Update interval"],
            order = 2,
            desc = L["Set the interval in which the HealGraph updates (in seconds)."],
            min = 0.01,
            max = 1.0,
            step = 0.01,
            bigStep = 0.02,            
            get = function() return HealGraph.db.profile.UpdateInterval end,
            set = "SetUpdateInterval",
        },
        createGraph = {
            type = "text",
            name = L["Create new graph"],
            desc = L["Creates a new graph with the passed id."],
            get = false,
            usage = L["<unique graph id>"],
            set = function(v) HealGraph:CreateGraph(v, v) end,
            order = -3
        },
        toggleIncDmg = {
            type = 'toggle',
            name = L["Show incoming damage (Experimental)"],
            order = 2,
            desc = L["Enables or disables the tracking and display of incoming damage. This is still very experimental and will probably produce some error. You can test it, but don't use it in any serious situation."],
            get = function() return HealGraph.db.profile.EnableIncDmg end,
            set = "EnableIncDamage",
        },
        --[[hideOutsideRaid = {
            type = 'toggle',
            name = L["Disable outside raid"],
            order = 2,
            desc = L["Toggles display of graphs outside of raids."],
            get = function() return HealGraph.db.profile.hideOutsideRaid end,
            set = "SetHideOutsideRaid",
        },]]
    },
}

---------------locals--------------------

local PlayerName = UnitName("player")

local Graph = HealGraphGraphClass




HealGraph:RegisterChatCommand("/healgraph", "/hg", options)

---update interval---
function HealGraph:EnableIncDamage(v)
    HealGraph.db.profile.EnableIncDmg = v
    
    if v then
        IncEvents:Enable()
    else
        IncEvents:Disable()
    end
end

function HealGraph:SetUpdateInterval(interval)
    self.db.profile.UpdateInterval = interval
    self:CancelScheduledEvent("HGUpdateGraph")
    self:ScheduleRepeatingEvent("HGUpdateGraph", "HGUpdateGraph", interval)
end

function HealGraph:CreateGraph(id, unit)
    self.db.profile.Graphs[id] = unit
    if not self.graphs[unit] then
        self.graphs[unit] = {}
    end
    self.graphs[unit][id] = {unit = unit, id = id}
    self.graphs[unit][id].graph = Graph:RestoreOrCreateGraph(id, unit)
    
    self.hasGraphs = true
    if unit == "target" then
        self.hasTargetGraph = true
    end
    if unit == "focus" then
        self.hasFocusGraph = true
    end
    
    local newOptions = self:GetOptionsFor(unit, id)
    
    options.args[id] = newOptions
end

function HealGraph:GetOptionsFor(unit, id)
    return {
        type='group',
        name = id,
        desc = "Options for a graph.",
        args = {
            timeSpan = {
                type = 'range',
                name = L["Time span"],
                order = 2,
                desc = L["Sets the time span the graph should display."],
                min = 0.5,
                max = 10.0,
                step = 0.25,
                bigStep = 0.5,            
                get = function() return HealGraph:GetRange(unit, id) end,
                set = function(v) return HealGraph:SetRange(unit, id, v) end,
            },
            pastSpan = {
                type = 'range',
                name = L["Time span for past"],
                order = 2,
                desc = L["Sets the time span lines which already happened should be displayed (the area left from the orange zero-line)."],
                min = 0.0,
                max = 2.0,
                step = 0.05,
                bigStep = 0.1,            
                get = function() return HealGraph:GetPastRange(unit, id) end,
                set = function(v) return HealGraph:SetPastRange(unit, id, v) end,
            },
            remove = {
                type = 'execute',
                name = L["Remove"],
                order = 2,
                desc = L["Removes the graph."],        
                func = function() return HealGraph:RemoveGraph(unit, id) end
            },
            createMark = {
                type = "text",
                name = L["Create mark"],
                desc = L["Creates a marking line at the specified time."],
                get = false,
                usage = L["<seconds>"],
                set = function(v) HealGraph:CreateMark(unit, id, v) end,
                order = -3
            },
            changeUnit = {
                type = "text",
                name = L["Change to unit"],
                desc = L["Changes the unit of the graph to the passed one. Can be any valid unitid or raidmember name. As well as \"healtarget\" for the unit currently healed."],
                get = false,
                usage = L["<unit name>"],
                set = function(v) HealGraph:ChangeGraphUnit(unit, id, v) end,
                order = -3
            },
            --[[hideOutsideRaid = {
                type = 'toggle',
                name = L["Hide outside raid"],
                order = 2,
                desc = L["Toggles hiding of this graph outside of raids."],
                get = function() return HealGraph:GetHideOutsideRaid(unit, id) end,
                set = function(v) HealGraph:SetHideOutsideRaid(unit, id, v) end,
            },]]
        },
        
    }
end

function HealGraph:CreateMark(unit, id, time)
    self.graphs[unit][id].graph:CreateMark(time)
end

function HealGraph:RemoveGraph(unit, id)
    self.db.profile.Graphs[id] = nil
    self.graphs[unit][id].graph:Disable()
    Graph:RemoveGraph(id)
    self.graphs[unit][id] = nil
    options.args[id] = nil
    
    self.hasGraphs = false
    self.hasTargetGraph = false
    self.hasFocusGraph = false
    for id, unit in pairs(self.db.profile.Graphs) do
        self.hasGraphs = true
        if unit == "target" then
            self.hasTargetGraph = true
        end
        if unit == "focus" then
            self.hasFocusGraph = true
        end
    end
end

--[[
function HealGraph:GetHideOutsideRaid(unit, id)
    return self.graphs[unit][id].graph:GetHideOutsideRaid()
end

function HealGraph:SetHideOutsideRaid(unit, id, v)
    self.graphs[unit][id].graph:SetHideOutsideRaid(v)
end
]]

function HealGraph:GetRange(unit, id)
    return self.graphs[unit][id].graph:GetTimeSpan()
end

function HealGraph:SetRange(unit, id, v)
    self.graphs[unit][id].graph:SetTimeSpan(v)
end

function HealGraph:GetPastRange(unit, id)
    return self.graphs[unit][id].graph:GetPastTimeSpan()
end

function HealGraph:SetPastRange(unit, id, v)
    self.graphs[unit][id].graph:SetPastTimeSpan(v)
end

function HealGraph:ChangeGraphUnit(unit, id, v)
    if unit == v then return end    
    
    if not self.graphs[v] then self.graphs[v] = {} end
    self.graphs[v][id] = self.graphs[unit][id]
    self.graphs[unit][id] = nil
    
    options.args[id] = self:GetOptionsFor(v, id)
    
    self.db.profile.Graphs[id] = v
    
    self.graphs[v][id].graph:ChangeUnit(v)
    
end
---------------initialisation--------------


function HealGraph:OnInitialize()
    self.heals = {}
    self.isHealing = nil

    self.expTime1 = {}
    self.expTime2 = {}
    self.start1 = {}
    self.start2 = {}
    self.wrong1 = 0
    self.wrong2 = 0

    Graph:SetDataBase(self.db.profile.GraphDB)
    
    if not self.db.profile.eventColors then
        self.db.profile.eventColors = {["swingMain"] = {r = 1, g = 1, b = 1, a = 1}, ["swingOff"] = {r = 1, g = 1, b = 1, a = 1}}
    end
    -----create graphs-----
    if not self.db.profile.Graphs then
        self.db.profile.Graphs = {["target"] = "target"}
    end
    
    
    self.graphs = {}
    for id, unit in pairs(self.db.profile.Graphs) do
        self.hasGraphs = true
        if unit == "target" then
            self.hasTargetGraph = true
        end
        if unit == "focus" then
            self.hasFocusGraph = true
        end
        self:Print("Creating graph for "..unit)
        self:CreateGraph(id, unit)
    end    
end

function HealGraph:OnEnable()
    self:EnableIncDamage(self.db.profile.EnableIncDmg)
    
    ---register events---
    
    if self.hasGraphs then
        self:RegisterEvent("UNIT_HEALTH", "UpdateHealth")
        self:RegisterEvent("UNIT_HEALTH_MAX", "UpdateHealth")
        self:RegisterEvent("UNIT_SPELLCAST_STOP", "SpellFailed")
        self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "SpellSucceded")
        self:RegisterEvent("UNIT_SPELLCAST_FAILED", "SpellFailed")
        self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "SpellFailed")
        self:ScheduleRepeatingEvent("HGUpdateGraph", "HGUpdateGraph", self.db.profile.UpdateInterval)    
        self:RegisterEvent("HGUpdateGraph")
    
        if self.hasTargetGraph then
            self:RegisterEvent("PLAYER_TARGET_CHANGED")
        end
        if self.hasFocusGraph then
            self:RegisterEvent("PLAYER_FOCUS_CHANGED")
        end
    end
    
    -- register callbacks
	HealComm.RegisterCallback(self, "HealComm_DirectHealStart")
	HealComm.RegisterCallback(self, "HealComm_DirectHealStop")
	HealComm.RegisterCallback(self, "HealComm_DirectHealDelayed")
	--HealComm.RegisterCallback(self, "HealComm_HealModifierUpdate")
    
    if not MobStats then
        IncEvents.RegisterCallback(self, "IncomingEvents_Started")
        IncEvents.RegisterCallback(self, "IncomingEvents_TimeChanged")
        IncEvents.RegisterCallback(self, "IncomingEvents_Stopped")
    else    
        MobStats.RegisterCallback(self, "IncomingEvents_Started")
        MobStats.RegisterCallback(self, "IncomingEvents_TimeChanged")
        MobStats.RegisterCallback(self, "IncomingEvents_Stopped")
    end
    ---show frames---
    self:EnableGraphs()
    
    ---update---
    self:Update()
end

function HealGraph:OnDisable()
    HealComm.UnregisterCallback(self, "HealComm_DirectHealStart")
	HealComm.UnregisterCallback(self, "HealComm_DirectHealStop")
	HealComm.UnregisterCallback(self, "HealComm_DirectHealDelayed")
	--HealComm.UnregisterCallback(self, "HealComm_HealModifierUpdate")
    
    if not MobStats then
        IncEvents.UnregisterCallback(self, "IncomingEvents_Started")
        IncEvents.UnregisterCallback(self, "IncomingEvents_TimeChanged")
        IncEvents.UnregisterCallback(self, "IncomingEvents_Stopped")
    else
        MobStats.UnregisterCallback(self, "IncomingEvents_Started")
        MobStats.UnregisterCallback(self, "IncomingEvents_TimeChanged")
        MobStats.UnregisterCallback(self, "IncomingEvents_Stopped")
    end
    
    ---hide frames---
    self:DisableGraphs()
    
    IncEvents:Disable()
end

function HealGraph:EnableGraphs()
    for unit, unitGraphs in pairs(self.graphs) do
        for id, graphData in pairs(unitGraphs) do
            graphData.graph:Enable()
        end
    end
end

function HealGraph:DisableGraphs()
    for unit, unitGraphs in pairs(self.graphs) do
        for id, graphData in pairs(unitGraphs) do
            graphData.graph:Disable()
        end
    end
end








-----------------event handling-----------------

--[[called whenever the current target of the player changes]]
function HealGraph:PLAYER_TARGET_CHANGED()
    --self:Print("target changed")
    self:RefreshGraphs("target")
end

--[[called whenever the current focus of the player changes]]
function HealGraph:PLAYER_FOCUS_CHANGED()
    self:RefreshGraphs("focus")
end

--[[This event is called in a specific interval and indicates
    to update the graphs]]
function HealGraph:HGUpdateGraph()
    self:Update()
end

function HealGraph:IncomingEvents_Started(event, eventName, sourceGUID, targetGUID, endTime, amount, prob)
    --self:Print("Incoming damage")    
    --self:Print(GetTime()..": IE - Swing: "..endTime..", "..(amount or "nil")..", "..prob.." \n-------------------- "..sourceGUID..", "..targetGUID)
    self.expTime2[sourceGUID] = endTime
	self:UpdateIncDamageFor(eventName, targetGUID, sourceGUID, endTime, amount, prob)
end

function HealGraph:IncomingEvents_TimeChanged(event, eventName, sourceGUID, targetGUID, endTime)
    --self:Print("Incoming damage")   
    --self:Print(GetTime()..": IE - !!!!CHANGE!!!!: "..endTime.." \n-------------------- "..sourceGUID..", "..targetGUID)
    self.expTime2[sourceGUID] = endTime
	self:UpdateIncDamageFor(eventName, targetGUID, sourceGUID, endTime)
end

function HealGraph:IncomingEvents_Stopped(event, eventName, sourceGUID, targetGUID, hit, amount)
    --self:Print("Incoming damage stopped")    
    --self:Print(GetTime()..": IE - Stop: "..(amount or "nil")..", "..(hit and "true" or "false").." \n-------------------- "..sourceGUID..", "..targetGUID)
	self:RemoveIncDamageFor(eventName, targetGUID, sourceGUID, amount)
end

function HealGraph:HealComm_DirectHealStart(event, healerFullName, healSize, endTime, ...)
    local ownHeal = false
    if not self.isHealing and healerFullName == PlayerName then
        self.isHealing = {}
        ownHeal = true
    end
    
	for i = 1, select("#", ...) do
		local targetFullName = select(i, ...)
		if ownHeal then
            tinsert(self.isHealing, targetFullName)
            if i == 1 then
                self:UpdateHealTargetGraphs(self.isHealing[1])
            end
        end
        self:UpdateIncHealFor(targetFullName, healerFullName, healSize, endTime)
	end
end

function HealGraph:HealComm_DirectHealStop(event, healerFullName, healSize, succeeded, ...)
	for i = 1, select("#", ...) do
		local targetFullName = select(i, ...)
        
        self:RemoveIncHealFor(targetFullName, healerFullName, healSize, succeeded)
	end
    
    --self:Print("heal stop")
    if self.isHealing and healerFullName == PlayerName then
        self.isHealing = nil
    end
end



function HealGraph:HealComm_DirectHealDelayed(event, healerFullName, healSize, endTime, ...)
	for i = 1, select("#", ...) do
		local targetFullName = select(i, ...)
        
        self:UpdateIncHealFor(targetFullName, healerFullName, healSize, endTime)
	end
end

function HealGraph:HealComm_HealModifierUpdate(event, unit, targetFullName, healModifier)
	--self:UpdateIncomingHeals(targetFullName)
end

function HealGraph:UpdateHealth(unitid)
    local fullName = self:FullUnitName(unitid)
    
    for unit, unitGraphs in pairs(self.graphs) do
        if fullName == self:FullUnitName(unit) then
            for id, graphData in pairs(unitGraphs) do
                graphData.graph:UpdateHealth()
            end
        end
    end
end

function HealGraph:SpellSucceded(unit)
    self:SpellStopped(unit, true)
end

function HealGraph:SpellFailed(unit)
    self:SpellStopped(unit, false)
end

function HealGraph:SpellStopped(unit, succeded)
    --[[local fullName = self:FullUnitName(unit)

    if self.isHealing and fullName == PlayerName then
        for _, targetFullName in ipairs(self.isHealing) do
            if succeded and self.heals[targetFullName] and self.heals[targetFullName][PlayerName] then
                self:RemoveIncHealFor(targetFullName, PlayerName, self.heals[targetFullName][PlayerName].amount, true)
            else
                self:RemoveIncHealFor(targetFullName, PlayerName, 0, false)
            end
        end
        self.isHealing = nil
    end]]
    
end



--------------actual functionality--------------

---general---

function HealGraph:GetColorForEvent(eventName)
    local color = self.db.profile.eventColors[eventName]
    
    if not color then
        local upperEvent = string.upper(eventName)
        local alpha = string.byte(upperEvent, string.len(upperEvent)) / 52.0 + 0.5
        color = {   r = (string.byte(upperEvent, 1) + string.byte(upperEvent, 2) + 120) * alpha,
                    g = (string.byte(upperEvent, 3) or 0 + string.byte(upperEvent, 4) or 0 + 180) * alpha,
                    r = (string.byte(upperEvent, 5) or 0 + string.byte(upperEvent, 6) or 0 + 180) * alpha,
                    a = 1}
                    
        self.db.profile.eventColors[eventName] = color
    end
    
    return color
end

function HealGraph:FullUnitName(unit)
    if self.isHealing and unit == "healtarget" then
        unit = self.isHealing[1]
    end
    
    local name, realm = UnitName(unit)
    
    if not name then return end
    
    if realm then
        return name.."-"..realm
    end
    
    return name
end

function HealGraph:UpdateHealTargetGraphs(targetFullName)
    local healTargetGraphs = self.graphs["healtarget"]
    if not healTargetGraphs then return end
    for id, graphData in pairs(healTargetGraphs) do
        graphData.graph:ChangeUnit(targetFullName)
    end
end

--[[Updates the heal graphs]]
function HealGraph:Update()
    for unit, unitGraphs in pairs(self.graphs) do
        for id, graphData in pairs(unitGraphs) do
            graphData.graph:UpdateTime()
        end
    end
end

function HealGraph:RefreshGraphs(unit)
    local fullName = self:FullUnitName(unit)
    
    
    local unitGraphs = self.graphs[unit]
    if not unitGraphs then
        return
    end
    
    for id, graphData in pairs(unitGraphs) do
        --self:Print("Refreshing "..id)
        local graph = graphData.graph
        graph:Clear()
        
        local targetHeals = (fullName and self.heals[fullName]) or nil
        if targetHeals then
            for healer, heal in pairs(targetHeals) do
                graph:AddOrChangeHeal(healer, heal.endTime, heal.amount, true)
            end
        end
        
        graph:UpdateUnit()
    end
end


---incoming heals tracker---
function HealGraph:RemoveIncHealFor(targetFullName, healerFullName, healSize, succeeded)
    local targetHeals = self.heals[targetFullName]
    local amount = nil
    if succeeded then
        amount = healSize
    end
    if targetHeals then
        targetHeals[healerFullName] = nil
        
        for unit, unitGraphs in pairs(self.graphs) do
            if targetFullName == self:FullUnitName(unit) then
                for id, graphData in pairs(unitGraphs) do
                    graphData.graph:RemoveHeal(healerFullName, amount)
                end
            end
        end
    end
end

function HealGraph:UpdateIncHealFor(targetFullName, healerFullName, healSize, endTime)
    local targetHeals = self.heals[targetFullName]
    if not targetHeals then
        self.heals[targetFullName] = {}
        targetHeals = self.heals[targetFullName]
    end
    local targetHealsFrom = targetHeals[healerFullName]
    if not targetHealsFrom then
        targetHeals[healerFullName] = {}
        targetHealsFrom = targetHeals[healerFullName]
    end
    targetHealsFrom.endTime = endTime
    targetHealsFrom.amount = healSize
    
    for unit, unitGraphs in pairs(self.graphs) do
        if targetFullName == self:FullUnitName(unit) then
            for id, graphData in pairs(unitGraphs) do
                graphData.graph:AddOrChangeHeal(healerFullName, endTime, healSize)
            end
        end
    end
end

function HealGraph:RemoveIncDamageFor(eventName, targetGUID, sourceGUID, amount)
    if amount then amount = -amount end
    local eventId = sourceGUID..eventName
    for unit, unitGraphs in pairs(self.graphs) do
        if targetGUID == UnitGUID(unit) then
            for id, graphData in pairs(unitGraphs) do
                graphData.graph:RemoveEvent(eventId, amount)
            end
        end
    end
end

function HealGraph:UpdateIncDamageFor(eventName, targetGUID, sourceGUID, endTime, amount, prob)   
    if amount then amount = -amount end
    local col = self:GetColorForEvent(eventName)
    col.a = max(prob or 1, 0.25)
    local eventId = sourceGUID..eventName
    for unit, unitGraphs in pairs(self.graphs) do      
        if targetGUID == UnitGUID(unit) then
            for id, graphData in pairs(unitGraphs) do
                
                graphData.graph:AddOrChangeLine(eventId, endTime, amount, col)
            end
        end
    end
end

