local MAJOR_VERSION = "LibIncomingDamage-1.0"
local MINOR_VERSION = tonumber(("$Revision: 72222 $"):match("%d+"))
 
local lib = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local playerName = UnitName('player')

local InBattlegroundOrArena
local InRaidOrParty

local refTime = nil
local refTimestamp = nil

IncomingDamageDB = {mobData = {}, spellData = {}}

local mobData = IncomingDamageDB.mobData

local expectedDamage = {}
local expectedDataBase = {}

-----------------
-- Event Frame --
-----------------

lib.EventFrame = lib.EventFrame or CreateFrame("Frame")
lib.EventFrame:SetScript("OnEvent", function (this, event, ...) lib[event](lib, ...) end)
lib.EventFrame:SetScript("OnUpdate", function (this, ...) lib.OnUpdate(lib, ...) end)
lib.EventFrame:UnregisterAllEvents()

-- Register Events
lib.EventFrame:RegisterEvent("PLAYER_ALIVE")
lib.EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

-- For keeping track of versions
lib.EventFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
lib.EventFrame:RegisterEvent("RAID_ROSTER_UPDATE")

-- Prune data at zone change
lib.EventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")



-------------------------------
-- Embed CallbackHandler-1.0 --
-------------------------------

lib.Callbacks = LibStub("CallbackHandler-1.0"):New(lib)


-----------------
-- Static Data --
-----------------

-- Cache of spells and heal sizes


---------------------------------
-- Frequently Accessed Globals --
---------------------------------

local bit_band = _G.bit.band


---------------
-- Utilities --
---------------

local function unitFullName(unit)
    local name, realm = UnitName(unit)
    if (realm and realm ~= "") then
        return name .. "-" .. realm
    else
        return name
    end
end

local function isPlayer(unitFlags)
    return bit_band(unitFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) == COMBATLOG_OBJECT_CONTROL_PLAYER
end

function lib:Print(msg)
    --DEFAULT_CHAT_FRAME:AddMessage("IncomingDamage: "..tostring(msg))
end

-----------------------------
-- Damage Data Management --
-----------------------------


----------------------
-- Public Functions --
----------------------


--------------------
-- Event Handlers --
--------------------


--self.Callbacks:Fire("HealComm_HealModifierUpdate", unit, targetName, newModifier);

local function getMobData(id)
    local mob = expectedDataBase[id]
    if not mob then
        expectedDataBase[id] = {swingTime = 2.0, count = 1, targetData = {}}
        mob = expectedDataBase[id]
    end
    
    return mob
end

function lib:SwingMissed(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,missType)
    if isPlayer(dstFlags) then 
        self:SwingOnPlayer(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, missType)
    elseif missType == "PARRY" then
        self:ParryOfMob(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, missType)
    end
end

function lib:ParryOfMob(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, missType)
    if not dstName then return end
    
    local expected = expectedDamage[dstGUID]
    
    if expected then
        self:Print("-----------------------------")
        self:Print("Parry of "..dstName)
        local left = expected - timestamp
        self:Print("Left: "..left)
        local mobID = tonumber(dstGUID:sub(-12,-7),16)
        local mob = getMobData(mobID)
        self:Print("Swingtime is: "..mob.swingTime.." with 0.4 of it "..mob.swingTime * 0.4)
        if left > mob.swingTime * 0.2 then
            left = math.max(mob.swingTime * 0.2, left - mob.swingTime * 0.4)
        end
        self:Print("New left: "..left)
        expectedDamage[dstGUID] = timestamp + left
        self:UpdateDamageFrom(dstGUID, expectedDamage[dstGUID] + refTime)
    end
end

function lib:SwingDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    if isPlayer(dstFlags) then 
        self:SwingOnPlayer(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, "HIT", amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    end
    
end

function lib:UnitDead(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)
    if expectedDamage[dstGUID] then 
        expectedDamage[dstGUID] = nil
        self:RemoveDamageFrom(dstGUID, nil)
        self.damageData[dstGUID] = nil
    end
end

function lib:UpdateDamageFrom(srcGUID, endTime)
    local incDamage = self.damageData[srcGUID]
    if not incDamage then return end
    
    incDamage.endTime = endTime
        
    self.Callbacks:Fire("IncomingDamage_SwingChangeTime", srcGUID, incDamage.targetGUID, endTime);
end

function lib:UpdateDamageOn(srcGUID, dstGUID, endTime, amount, prob)
    local incDamage = self.damageData[srcGUID]
    if not incDamage then
        self.damageData[srcGUID] = {targetGUIDs = {}}
        incDamage = self.damageData[srcGUID]
    end
    
    incDamage.endTime = endTime
    incDamage.targetGUID = dstGUID
    incDamage.amount = amount
    incDamage.prob = prob
        
    self.Callbacks:Fire("IncomingDamage_Swing", srcGUID, dstGUID, endTime, amount, prob);
end

function lib:RemoveDamageFrom(srcGUID, sucAmount)
    local incDamage = self.damageData[srcGUID]
    if not incDamage then return end
    self.Callbacks:Fire("IncomingDamage_SwingStop", srcGUID, incDamage.targetGUID, sucAmount);
    incDamage.targetGUIDs = nil
    incDamage.endTime = nil
end

function lib:AddHitToMob(mob, dstGUID, damage)
    local hitData = mob.targetData[dstGUID]
    
    if not hitData then
        mob.targetData[dstGUID] = {count = 0, hits = 0, damage = 0}
        hitData = mob.targetData[dstGUID]
    end
    
    hitData.count = hitData.count + 1
    hitData.hits = hitData.hits + 1
    
    if hitData.hits < 100 then
        hitData.damage = hitData.damage + damage
    else
        if hitData.hits == 100 then
            hitData.damage = hitData.damage / hitData.hits
        end
        
        hitData.damage = hitData.damage + (damage - hitData.damage) / hitData.hits
    end
end

function lib:AddMissToMob(mob, dstGUID)
    local hitData = mob.targetData[dstGUID]
    
    if not hitData then
        mob.targetData[dstGUID] = {count = 0, hits = 0, damage = 0}
        hitData = mob.targetData[dstGUID]
    end
    
    hitData.count = hitData.count + 1
end

function lib:GetHitData(mob, dstGUID)
    local hitData = mob.targetData[dstGUID]
    
    local damage = 0
    if hitData.hits > 0 then
        if hitData.hits < 100 then
            damage = hitData.damage / hitData.hits
        else
            damage = hitData.damage
        end
    end
    return hitData.hits / hitData.count, damage
end

function lib:SwingOnPlayer(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, hitType, amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    if not srcName then return end
    
    local expected = expectedDamage[srcGUID]
    local mobID = tonumber(srcGUID:sub(-12,-7),16)
    local mobData = getMobData(mobID)
    
    
    self:RemoveDamageFrom(srcGUID, amount)
    
    if expected then
        local delta = timestamp - expected
        self:Print("-----------------------------")
        self:Print("Delta was: "..delta)
        if delta < mobData.swingTime * 0.5 then
            self:Print("Previous swingtime: "..mobData.swingTime)
            mobData.swingTime = mobData.swingTime + delta / mobData.count
            mobData.count = mobData.count + 1
            self:Print("Updated swingtime to: "..mobData.swingTime)
        end
    end
    
    if hitType == "HIT" then
        self:AddHitToMob(mobData, dstGUID, amount + (absorbed or 0))
    else
        if hitType ~= "ABSORB" then
            self:AddMissToMob(mobData, dstGUID)
        end
    end
    
    local prob, amount = self:GetHitData(mobData, dstGUID)
    self:Print("Damage: "..amount)
    self:Print("Prob.: "..prob)
    
    expectedDamage[srcGUID] = timestamp + mobData.swingTime
    self:UpdateDamageOn(srcGUID, dstGUID, expectedDamage[srcGUID] + refTime, amount, prob)
end

local EventParse =
{
	["SWING_DAMAGE"] = lib.SwingDamage, 
	["RANGE_DAMAGE"] = lib.SwingDamage, 
	["SWING_MISSED"] = lib.SwingMissed, 
	["RANGE_MISSED"] = lib.SwingMissed,
    ["UNIT_DIED"] = lib.UnitDead,
}


function lib:COMBAT_LOG_EVENT_UNFILTERED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
    
    refTime = GetTime() - timestamp
	local parsefunc = EventParse[eventtype]
	
	if parsefunc then
		parsefunc(self, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	else
		--self:Print((srcName or "<nil>")..": Unknown combat log event type: "..eventtype)
	end
end

local elapsedTimer = 0

function lib:OnUpdate(elapsed)
    elapsedTimer = elapsedTimer + elapsed
    if elapsedTimer > 0.25 then
        for srcGUID, inc in pairs(self.damageData) do
            if inc.endTime then
                local mobID = tonumber(srcGUID:sub(-12,-7),16)
                local mobData = getMobData(mobID)
                
                local delta = GetTime() - inc.endTime
                if delta > mobData.swingTime * 0.5 then
                    self:RemoveDamageFrom(srcGUID, nil)
                end
            end
        end
        elapsedTimer = 0
    end
end

function lib:PLAYER_ALIVE()
    -- This event is only fired at initial login, not at reloadui or load-on-demand loading.
    -- The initialisation is triggered again, since none of the initialisation had any effect
    -- prior to this event firing (no messages sent and InBattlegroundOrArena and InRaidOrParty
    -- are probably not correctly initialised).
    lib:Initialise();
end

function lib:PLAYER_ENTERING_WORLD()
    --cleanup
    self:Print("entering")
    self.damageData = {}
    expectedDamage = {}
    expectedDataBase = {}
end

function lib:PARTY_MEMBERS_CHANGED()
    InRaidOrParty = (GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0)

end

function lib:RAID_ROSTER_UPDATE()
    self:PARTY_MEMBERS_CHANGED()
end

function lib:Initialise()
    self.damageData = {}
    local it = select(2, IsInInstance())
    InBattlegroundOrArena = (it == "pvp") or (it == "arena")

    InRaidOrParty = (GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0)

end

lib:Initialise()
